/* Copyright 2025, Michele "King_DuckZ" Santullo
 * This file is part of ducktorrent.
 *
 * Ducktorrent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ducktorrent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ducktorrent.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ducktorrent/torrent_read.hpp"
#include "ducktorrent_config.h"
#include "git_version.h"

#include <iostream>
#include <fstream>
#include <string>
#include <algorithm>
#include <cstdint>
#include <sha1/sha1.hpp>
#include <sha1/BufferSource.hpp>
#include <climits>
#include <tclap/CmdLine.h>

namespace {
struct CmdOutPrinter : TCLAP::StdOutput {
	void version (TCLAP::CmdLineInterface& c) override {
		using duck::g_project_name;
		using duck::g_project_version;
		using duck::g_proj_git_sha1;

		std::cout << '\n';

		std::cout << g_project_name << "  Copyright (C) 2025  Michele \"King_DuckZ\" Santullo\n" <<
			"This program comes with ABSOLUTELY NO WARRANTY;\n" <<
			"This is free software, and you are welcome to redistribute it\n" <<
			"under certain conditions; see the source for copying conditions.\n";

		std::cout << '\n' << g_project_name << "  version: " <<
			g_project_version << '\n';
		if constexpr (g_proj_git_sha1[0] != '\0') {
			std::cout << "build commit: " << g_proj_git_sha1 << '\n';
		}
		std::cout << '\n';
	}
};

SHA1::MessageDigest to_hash160_digest (const std::array<std::uint32_t, 5>& arr) {
	SHA1::MessageDigest hash_found;
	std::copy_n(arr.begin(), 5, hash_found.hash);
	return hash_found;
}
} //unnamed namespace

int main(int argc, const char* argv[]) {
	std::string torrent_path;

	try {
		CmdOutPrinter custom_printer;

		TCLAP::CmdLine cmd("How to use " PROJECT_NAME, ' ', duck::g_project_version);
		cmd.setOutput(&custom_printer);

		TCLAP::ValueArg<std::string> dir_par("p", "path",
			"Path to the location where torrent files are located",
			false, "", "path");
		cmd.add(dir_par);

		TCLAP::UnlabeledValueArg<std::string> torrent_par("torrent_file",
			"Path of the torrent file to load",
			true, "", "path", false);
		cmd.add(torrent_par);

		cmd.parse(argc, argv);

		torrent_path = torrent_par.getValue();
	}
	catch (TCLAP::ArgException& e) {
		std::cerr << "Error: " << e.error() << " for parameter " << e.argId() << std::endl;
		return 2;
	}

	duck::TorrentRead torrent(torrent_path, "");
	std::cout << "Loaded file into string of size " << torrent.raw_data_size() << '\n';

	std::cout << "Torrent name: " << torrent.read_name() << '\n';
	std::cout << "Created by: " << torrent.read_created_by() << '\n';
	std::cout << "Creation date: " << torrent.read_creation_date() << '\n';
	std::cout << "Comment: " << torrent.read_comment() << '\n';
	std::cout << "Announce: " << torrent.read_announce() << '\n';
	{
		std::size_t z = 0;
		for (const std::vector<std::string_view>& sublist : torrent.read_announce_list()) {
			std::cout << '\t' << ++z << ": < ";
			for (std::string_view announce : sublist) {
				std::cout << '"' << announce << "\", ";
			}
			std::cout << " >\n";
		}
	}
	std::cout << "Is private: " << std::boolalpha << torrent.read_is_private() << '\n';

	torrent.print(std::cout);

	auto hashes = duck::group_torrent_hashes(torrent);
	std::cout << "Got " << hashes.size() << " hashes\n";

	const auto file_count = torrent.read_file_count();
	if (file_count)
		std::cout << "Input has " << file_count << " file entries\n";
	else
		std::cout << "Input seems to contain a single file only\n";

	const std::size_t piece_length = torrent.read_piece_length();
	if (0 == file_count) {
		std::string file_name{torrent.read_name()};
		std::cout << "Found file name \"" << file_name << "\"\n";

		std::ifstream istream(file_name, std::ios::in|std::ios::binary);

		if (istream.is_open()) {
			std::vector<std::uint8_t> buff(piece_length);
			std::size_t hash_index = 0;
			std::size_t match_count = 0;

			while (istream.read(reinterpret_cast<char*>(buff.data()), piece_length).gcount() > 0) {
				SHA1::BufferSource buff_source(buff.data(), istream.gcount() * CHAR_BIT);
				SHA1::MessageDigest calculated_hash = SHA1::computeFromSource(buff_source);
				SHA1::MessageDigest stored_hash = to_hash160_digest(hashes[hash_index++]);
				std::cout << stored_hash.toHexString() << "  " <<
					calculated_hash.toHexString() << "  " << file_name << '\n';
				if (stored_hash == calculated_hash)
					++match_count;
			}

			std::cout << "Hash verified " << match_count << '/' << hashes.size() << '\n';
		}
		else {
			std::cout << "Error: opening \"" << file_name << "\" failed\n";
		}
	}
	else {
		std::size_t hash_index = 0;
		std::size_t match_count = 0;
		std::size_t read_size = 0;
		std::vector<std::uint8_t> buff;
		for (std::int_fast32_t z = 0; z < file_count; ++z) {
			std::string curr_path = torrent.read_joint_file_path(z, '/');
			std::cout << "path index " << z << '\t' << curr_path << '\n';

			std::ifstream istream(curr_path, std::ios::in|std::ios::binary);
			if (istream.is_open()) {
				buff.resize(piece_length);
				while (istream.read(reinterpret_cast<char*>(buff.data() + read_size), piece_length - read_size).gcount() > 0) {
					read_size += istream.gcount();
					if (piece_length == read_size or file_count == z + 1) {
						SHA1::BufferSource buff_source(buff.data(), read_size * CHAR_BIT);
						SHA1::MessageDigest calculated_hash = SHA1::computeFromSource(buff_source);
						SHA1::MessageDigest stored_hash = to_hash160_digest(hashes[hash_index++]);
						std::cout << stored_hash.toHexString() << "  " <<
							calculated_hash.toHexString() << "  " << curr_path << '\n';
						if (stored_hash == calculated_hash)
							++match_count;
						read_size = 0;
					}
				}
			}
			else {
				std::cout << "Unable to open " << curr_path << ", bailing out of hash verification\n";
				break;
			}
		}
		if (hash_index == hashes.size()) {
			std::cout << "Hash verified " << match_count << '/' << hashes.size() << '\n';
		}
	}
}
