/* Copyright 2025, Michele "King_DuckZ" Santullo
 * This file is part of ducktorrent.
 *
 * Ducktorrent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ducktorrent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ducktorrent.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ducktorrent/torrent_read.h"
#include "ducktorrent/torrent_read.hpp"

#include <memory>
#include <ciso646>
#include <string_view>
#include <iostream>
#include <string>
#include <vector>

extern "C"
struct dt_torrent {
	dt_torrent (std::string_view path) :
		reader(path, "")
	{ }

	duck::TorrentRead reader;
	std::string str_buff;
	std::vector<dt_string> vec_buff;
	std::vector<std::vector<dt_string>> announce_list_buff;
};

extern "C"
struct dt_torrent* dt_make_reader_from_path (dt_string path) {
	try {
		std::string_view path_sv(path.str, path.len);
		return new dt_torrent(path_sv);
	}
	catch (...) {
		return nullptr;
	}
}

namespace {
//Just a convenience wrapper because I found I was copy-pasting code to fetch
//arrays from TorrentRead just because each C function would call into a
//different method. So here I take the method as a template parameter and no
//more copy-paste
dt_string_list read_string_list (struct dt_torrent* t, auto method, auto&&... args) {
	try {
		auto vec = (t->reader.*method)(std::forward<decltype(args)>(args)...);
		t->vec_buff.clear();
		t->vec_buff.reserve(vec.size());
		for (std::string_view s : vec) {
			t->vec_buff.emplace_back(s.data(), s.size());
		}
		return {t->vec_buff.data(), t->vec_buff.size()};
	}
	catch (...) {
		return {nullptr, 0};
	}
}
} //unnamed namespace

extern "C"
void dt_free_torrent_read (struct dt_torrent* t) {
	delete t;
}

extern "C"
void dt_print_to_stdout(struct dt_torrent* t) {
	try {
		t->reader.print(std::cout);
	}
	catch (...) {
	}
}

extern "C"
size_t dt_read_piece_length(struct dt_torrent* t) {
	try {
		return t->reader.read_piece_length();
	}
	catch (...) {
		return -1;
	}
}

extern "C"
int_fast32_t dt_read_file_count(struct dt_torrent* t) {
	try {
		return t->reader.read_file_count();
	}
	catch (...) {
		return -1;
	}
}

extern "C"
dt_string_list dt_read_file_path(struct dt_torrent* t, size_t index) {
	return read_string_list(t, &duck::TorrentRead::read_file_path, index);
}

extern "C"
dt_string_list dt_read_url_list(struct dt_torrent* t) {
	return read_string_list(t, &duck::TorrentRead::read_url_list);
}

extern "C"
dt_string dt_read_joint_file_path(struct dt_torrent* t, size_t index, char sep){
	try {
		t->str_buff = t->reader.read_joint_file_path(index, sep);
		return {t->str_buff.c_str(), t->str_buff.size()};
	}
	catch (...) {
		return {nullptr, 0};
	}
}

extern "C"
size_t dt_read_file_size(struct dt_torrent* t, size_t index) {
	try {
		return t->reader.read_file_size(index);
	}
	catch (...) {
		return -1;
	}
}

extern "C"
int dt_read_is_private(struct dt_torrent* t) {
	try {
		return static_cast<int>(t->reader.read_is_private());
	}
	catch (...) {
		return -1;
	}
}

extern "C"
dt_string dt_read_comment(struct dt_torrent* t) {
	try {
		std::string_view comment = t->reader.read_comment();
		return dt_string{comment.data(), comment.size()};
	}
	catch (...) {
		return {nullptr, 0};
	}
}

extern "C"
dt_string dt_read_hashes(struct dt_torrent* t) {
	try {
		std::string_view hashes = t->reader.read_hashes();
		return dt_string{hashes.data(), hashes.size()};
	}
	catch (...) {
		return {nullptr, 0};
	}
}

extern "C"
dt_string dt_read_created_by(struct dt_torrent* t) {
	try {
		std::string_view created = t->reader.read_created_by();
		return dt_string{created.data(), created.size()};
	}
	catch (...) {
		return {nullptr, 0};
	}
}

extern "C"
dt_string dt_read_announce(struct dt_torrent* t) {
	try {
		std::string_view announce = t->reader.read_announce();
		return dt_string{announce.data(), announce.size()};
	}
	catch (...) {
		return {nullptr, 0};
	}
}

extern "C"
size_t dt_read_announce_list_size(struct dt_torrent* t) {
	try {
		return t->reader.read_announce_list_size();
	}
	catch (...) {
		return -1;
	}
}

extern "C"
dt_string_list dt_read_announce_list(struct dt_torrent* t, size_t index) {
	try {
		if (t->announce_list_buff.empty()) {
			t->announce_list_buff.reserve(t->reader.read_announce_list_size());
			for (const auto& sublist : t->reader.read_announce_list()) {
				t->announce_list_buff.emplace_back();
				for (std::string_view item : sublist) {
					t->announce_list_buff.back().emplace_back(item.data(), item.size());
				}
			}
		}
		return {t->announce_list_buff[index].data(), t->announce_list_buff[index].size()};
	}
	catch (...) {
		return {nullptr, 0};
	}
}

extern "C"
time_t dt_read_creation_date(struct dt_torrent* t) {
	try {
		return t->reader.read_creation_date();
	}
	catch (...) {
		return {};
	}
}
