/* Copyright 2025, Michele "King_DuckZ" Santullo
 * This file is part of ducktorrent.
 *
 * Ducktorrent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ducktorrent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ducktorrent.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

//c wrapper for torrent_read.hpp

#if defined(__cplusplus)
extern "C" {
#endif

#include <stddef.h>
#include <stdint.h>
#include <time.h>

/* Strings and lists returned by functions in this file should be considered
 * temporary and might become invalid on any subsequent function call. There is
 * no thread safety mechanism so you should protect calls with your own mutex.
 * With that said, if you're interested in avoiding unnecessary memory
 * allocations and copies, there are cases where the above is not true. Rather
 * than making a list of functions the whose output is guaranteed to not change,
 * the rule for figuring out what might not need copying is the following:
 * wrapper functions around C++ methods that return an std::string_view do not
 * need to be copied. That is because the string returned by the C++ side is
 * already a reference into the in-memory torrent file buffer. This however
 * means that these particular strings are *not* null-terminated.
 * As these functions do not write to the C wrapper's internal string cache and
 * they only read from the parsed tree, it is also safe to call them in a
 * multithreaded context.
 * One more exception to the above rule is the special case of
 * dt_read_announce_list(), since it returs a list of lists this C wrapper
 * caches the whole thing on the first call and subsequent calls will not hit
 * the C++ side again. As said, the list side of things is cached in the C
 * wrapper, and since the inner type returned by the C++ method is
 * std::string_view this means the strings won't change either, therefore it is
 * safe to not copy these items either for as long as the dt_torrent context
 * stays alive.
 */

typedef struct {
	const char* str;
	size_t len;
} dt_string;

typedef struct {
	const dt_string* lst;
	size_t len;
} dt_string_list;

/*struct dt_torrent_read* dt_make_reader_from_data (const char* raw_data, size_t raw_data_size);*/
struct dt_torrent* dt_make_reader_from_path (dt_string path);
void dt_free_torrent_read (struct dt_torrent*);

void dt_print_to_stdout(struct dt_torrent*);
size_t dt_read_piece_length(struct dt_torrent*);
int_fast32_t dt_read_file_count(struct dt_torrent*);
dt_string_list dt_read_file_path(struct dt_torrent*, size_t index);
dt_string_list dt_read_url_list(struct dt_torrent*);
dt_string dt_read_joint_file_path(struct dt_torrent*, size_t index, char sep);
size_t dt_read_file_size(struct dt_torrent*, size_t index);
int dt_read_is_private(struct dt_torrent*);
dt_string dt_read_comment(struct dt_torrent*);
dt_string dt_read_hashes(struct dt_torrent*);
dt_string dt_read_created_by(struct dt_torrent*);
dt_string dt_read_announce(struct dt_torrent*);
size_t dt_read_announce_list_size(struct dt_torrent*);
dt_string_list dt_read_announce_list(struct dt_torrent*, size_t index);
time_t dt_read_creation_date(struct dt_torrent*);

#if defined(__cplusplus)
}
#endif
