/* Copyright 2025, Michele "King_DuckZ" Santullo
 * This file is part of ducktorrent.
 *
 * Ducktorrent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ducktorrent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ducktorrent.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ducktorrent/torrent_read.hpp"
#include "ducktorrent/parser.hpp"
#include "visitors/debug_visitor.hpp"
#include "visitors/find_t_visitor.hpp"
#include <utility>
#include <fstream>
#include <system_error>
#include <boost/variant/apply_visitor.hpp>
#include <ciso646>
#include <cassert>
#include <algorithm>
#include <ranges>
#include <numeric>

namespace duck {
namespace {
std::string load_file (const std::filesystem::path& path) {
	std::ifstream ifs(path);
	if (ifs.is_open()) {
		ifs >> std::noskipws;

		return std::string(
			std::istreambuf_iterator<char>(ifs),
			std::istreambuf_iterator<char>()
		);
	}
	else {
		throw std::filesystem::filesystem_error(
			"Error reading " + path.string(),
			std::make_error_code(std::errc::no_such_file_or_directory)
		);
	}
}

template <std::size_t OutS, std::size_t PrefS, std::size_t PostS>
std::string_view build_fast_search_path (
	const char (&prefix)[PrefS],
	const char (&postfix)[PostS],
	std::size_t index,
	char (&out_buff)[OutS],
	std::string& emergency_buff
) {
	static_assert(PrefS >= 1 and PostS >= 1, "Expected null-terminated inputs");

	constexpr std::size_t prefix_size = PrefS - 1;
	constexpr std::size_t postfix_size = PostS - 1;
	constexpr std::size_t output_max_size = OutS;

	std::string_view retval;
	auto num = std::to_string(index); //short string optimisation should prevent allocation
	const std::size_t real_out_size = prefix_size + num.size() + postfix_size;

	if (real_out_size <= output_max_size) {
		//Avoid allocations, just because
		std::copy_n(prefix, prefix_size, out_buff);
		std::copy_n(num.data(), num.size(), out_buff + prefix_size);
		std::copy_n(postfix, postfix_size, out_buff + prefix_size + num.size());
		retval = std::string_view(out_buff, prefix_size + num.size() + postfix_size);
	}
	else {
		//too long, I yield, allocate away!!
		emergency_buff.reserve(real_out_size);
		std::string_view prefix_sv{prefix, prefix_size};
		emergency_buff += prefix_sv;
		emergency_buff += num;
		std::string_view postfix_sv{postfix, postfix_size};
		emergency_buff += postfix_sv;
		retval = emergency_buff;
	}

	return retval;
}

std::vector<std::string_view> grab_all_strings_in_list (const TorrentValue& node) {
	const auto path_size = find_int<std::int_fast32_t>("/[[size]]", node);
	std::vector<std::string_view> retval;
	retval.reserve(path_size);

	{
		constexpr char prefix[] = "/[[";
		constexpr char postfix[] = "]]";
		constexpr std::size_t prefix_size = sizeof(prefix) / sizeof(prefix[0]) - 1u;
		constexpr std::size_t postfix_size = sizeof(postfix) / sizeof(postfix[0]) - 1u;
		char buff[4 + prefix_size + postfix_size];
		std::string buff2;

		for (std::int_fast32_t z = 0; z < path_size; ++z) {
			std::string_view piece = find_string(
				build_fast_search_path(prefix, postfix, z, buff, buff2),
				node
			);
			retval.push_back(piece);
		}
	}

	return retval;
}
} //unnamed namespace

TorrentRead::TorrentRead (std::string_view torrent_path, std::filesystem::path workdir) :
	m_workdir(std::move(workdir)),
	m_torrent_path(torrent_path),
	m_raw_torrent(load_file(m_torrent_path)),
	m_parsed_values(parse_torrent(m_raw_torrent)),
	m_cached_info_files(nullptr)
{
}

TorrentRead::~TorrentRead() noexcept = default;

void TorrentRead::print (std::ostream& out) const {
	DebugVisitor visitor(out);
	for (const TorrentValue& value : m_parsed_values) {
		boost::apply_visitor(visitor, value);
	}
}

std::size_t TorrentRead::raw_data_size() const {
	return m_raw_torrent.size();
}

const std::vector<TorrentValue>& TorrentRead::parsed_values() const {
	return m_parsed_values;
}

std::string_view TorrentRead::read_name() const {
	return find_string("/info/name", m_parsed_values);
}

std::size_t TorrentRead::read_piece_length() const {
	return find_int<std::size_t>("/info/piece length", m_parsed_values);
}

std::int_fast32_t TorrentRead::read_file_count() const {
	return find_int<std::int_fast32_t>("/[[size]]", cached_info_files());
}

std::vector<std::string_view> TorrentRead::read_file_path(std::size_t index) const {
	const TorrentValue* path_variant;

	{
		constexpr char prefix[] = "/[[";
		constexpr char postfix[] = "]]/path";
		constexpr std::size_t prefix_size = sizeof(prefix) / sizeof(prefix[0]) - 1u;
		constexpr std::size_t postfix_size = sizeof(postfix) / sizeof(postfix[0]) - 1u;

		//see comment in read_file_size()
		//in this case a search should look like "/[[1234]]/path"
		char buff[4 + prefix_size + postfix_size];
		std::string buff2;

		auto found = find_variant(
			build_fast_search_path(prefix, postfix, index, buff, buff2),
			cached_info_files()
		);
		if (not found)
			return {};

		path_variant = &found->get();
	}

	return grab_all_strings_in_list(*path_variant);
}

std::vector<std::string_view> TorrentRead::read_url_list() const {
	auto found = find_variant("/url-list", m_parsed_values);
	if (not found)
		return {};

	return grab_all_strings_in_list(*found);
}

std::string TorrentRead::read_joint_file_path(std::size_t index, char sep) const {
	auto pieces = this->read_file_path(index);
	const std::size_t out_size = std::accumulate(pieces.cbegin(), pieces.cend(), 0u,
		[](std::size_t a, const std::string_view& b) {return a+b.size();}
	);

	std::string retval;
	retval.reserve(out_size + std::max<std::size_t>(pieces.size(), 1u) - 1u);

	std::ranges::copy(pieces | std::views::join_with(sep), std::back_inserter(retval));
	return retval;
}

std::size_t TorrentRead::read_file_size (std::size_t index) const {
	constexpr char prefix[] = "/[[";
	constexpr char postfix[] = "]]/length";
	constexpr std::size_t prefix_size = sizeof(prefix) / sizeof(prefix[0]) - 1u;
	constexpr std::size_t postfix_size = sizeof(postfix) / sizeof(postfix[0]) - 1u;

	//accomodate paths with indices up to 9999 without nullchar
	//ie: "/[[1234]]/length" like this, if this comment doesn't go out of sync
	//with the code the size on the stack should be 16
	char buff[4 + prefix_size + postfix_size];
	std::string buff2;

	return find_int<std::size_t>(
		build_fast_search_path(prefix, postfix, index, buff, buff2),
		cached_info_files()
	);
}

bool TorrentRead::read_is_private() const {
	const auto val = find_int<unsigned int>("/info/private", m_parsed_values);
	return (val ? true : false);
}

std::string_view TorrentRead::read_comment() const {
	return find_string("/comment", m_parsed_values);
}

std::string_view TorrentRead::read_hashes() const {
	return find_string("/info/pieces", m_parsed_values);
}

std::string_view TorrentRead::read_created_by() const {
	return find_string("/created by", m_parsed_values);
}

std::string_view TorrentRead::read_announce() const {
	return find_string("/announce", m_parsed_values);
}

std::size_t TorrentRead::read_announce_list_size() const {
	return find_int<std::size_t>("/announce/[[size]]", m_parsed_values);
}

std::vector<std::vector<std::string_view>> TorrentRead::read_announce_list() const {
	typedef boost::spirit::x3::forward_ast<std::vector<TorrentValue>> TorrentListType;

	//see https://bittorrent.org/beps/bep_0012.html it's a list of lists
	auto list_variant = find_variant("/announce-list", m_parsed_values);
	if (list_variant) {
		auto* node_list = boost::get<TorrentListType>(&list_variant->get());
		std::vector<std::vector<std::string_view>> retval;
		if (node_list) {
			const auto list_size = static_cast<std::int_fast32_t>(node_list->get().size());
			retval.reserve(list_size);
			for (std::int_fast32_t z = 0; z < list_size; ++z) {
				retval.push_back(grab_all_strings_in_list(node_list->get()[z]));
			}
		}
		return retval;
	}
	else {
		//from my tests this node is optional
		return {};
	}
}

std::time_t TorrentRead::read_creation_date() const {
	return find_int<std::time_t>("/creation date", m_parsed_values);
}

const TorrentValue& TorrentRead::cached_info_files() const {
	constexpr char info_files[] = "/info/files";
	if (not m_cached_info_files) {
		auto found = find_variant(info_files, m_parsed_values);
		if (found) {
			m_cached_info_files = &found->get();
		}
	}

	if (not m_cached_info_files) {
		throw std::runtime_error(
			std::string{"Node \""} + info_files + "\" not found in " + m_torrent_path.string()
		);
	}

	assert(m_cached_info_files);
	return *m_cached_info_files;
}

std::vector<std::array<std::uint32_t, 5>> group_torrent_hashes (std::string_view hashes) {
	typedef std::array<std::uint32_t, 5> array_type;

	if (hashes.empty())
		return {};

	constexpr std::size_t hash_size = 20;
	constexpr std::size_t uint_count = std::tuple_size<array_type>();
	static_assert(sizeof(array_type::value_type) * uint_count == hash_size);

	if (hashes.size() % hash_size != 0)
		throw std::runtime_error("Bad pieces array size " + std::to_string(hashes.size()));

	std::vector<array_type> retval(hashes.size() / hash_size);
	const char* src = hashes.data();
	for (std::size_t z = 0; z < retval.size(); ++z, src+=hash_size) {
		//std::copy_n(src, hash_size, reinterpret_cast<char*>(retval[z].data()));
		for (std::size_t n = 0; n < uint_count; ++n) {
			char* const out_uint = reinterpret_cast<char*>(retval[z].data() + n);
			for (std::size_t u = 0; u < sizeof(std::uint32_t); ++u) {
				out_uint[u] = src[n * sizeof(std::uint32_t) + sizeof(std::uint32_t) - u - 1];
			}
		}
	}
	return retval;
}

std::vector<std::array<std::uint32_t, 5>> group_torrent_hashes (const TorrentRead& torrent) {
	return group_torrent_hashes(torrent.read_hashes());
}
} //namespace duck
