/* Copyright 2025, Michele "King_DuckZ" Santullo
 * This file is part of ducktorrent.
 *
 * Ducktorrent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ducktorrent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ducktorrent.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "find_t_visitor.hpp"
#include "split.hpp"

#include <string>
#include <regex>
#include <cassert>

namespace duck::detail {
namespace {
template <typename T>
void assert_same_value (const TorrentValue& var, const T& val) {
	assert(boost::get<T>(&var) != nullptr and *boost::get<T>(&var) == val);
}
} //unnamed namespace

template <typename T>
FindTVisitor<T>::FindTVisitor (std::string_view search_path) {
	auto split = dincore::split(search_path, '/', true, true);
	m_search.reserve(split.size());
	for (const auto& itm : split) {
		m_search.emplace_back(itm.data(), itm.size());
	}
}

template <typename T>
void FindTVisitor<T>::reset (const TorrentValue* curr_variant) {
	m_search_error = false;
	m_search_match_count = 0;

	m_curr_variant = curr_variant;
	assert(m_curr_variant);
}

template <typename T>
bool FindTVisitor<T>::has_error() const noexcept {
	return m_search_error;
}

template <typename T>
auto FindTVisitor<T>::operator() (TorrentIntType value) -> TRet {
	return scalar_operator_impl(value);
}

template <typename T>
auto FindTVisitor<T>::operator() (const TorrentStringType& value) -> TRet {
	return scalar_operator_impl(value);
}

template <typename T>
auto FindTVisitor<T>::operator() (const std::vector<TorrentValue>& values) -> TRet {
	if (is_path_found()) {
		if constexpr (std::is_same<TorrentValue, T>::value)
			return std::cref(*m_curr_variant);
		else
			return {};
	}

	if constexpr (std::is_same<TorrentIntType, T>::value) {
		if (m_search[m_search_match_count] == "[[size]]")
			return static_cast<TorrentIntType>(values.size());
	}

	std::regex reg_index(R"(^\[\[(\d+)\]\]$)");
	std::smatch match;
	const bool matched = std::regex_match(m_search[m_search_match_count], match, reg_index);
	if (matched) {
		++m_search_match_count;
		const auto index = static_cast<std::size_t>(std::stoul(match[1]));
		if (index < values.size()) {
			m_curr_variant = &values[index];
			return boost::apply_visitor(*this, values[index]);
		}
	}
	return {};
}

template <typename T>
auto FindTVisitor<T>::operator() (const std::map<TorrentStringType, TorrentValue>& value) -> TRet {
	if (is_path_found()) {
		if constexpr (std::is_same<TorrentValue, T>::value) {
			return std::cref(*m_curr_variant);
		}
		else {
			m_search_error = true;
			return {};
		}
	}

	auto it_found = value.find(m_search[m_search_match_count]);
	if (value.end() == it_found)
		return {};

	++m_search_match_count;
	m_curr_variant = &it_found->second;
	return boost::apply_visitor(*this, it_found->second);
}

template <typename T>
template <typename U>
auto FindTVisitor<T>::scalar_operator_impl (const U& value) -> TRet {
	if constexpr (std::is_same<TorrentValue, T>::value) {
		assert_same_value(*m_curr_variant, value);
		return std::cref(*m_curr_variant);
	}
	else if constexpr (std::is_same<U, T>::value) {
		if (is_path_found())
			return value;
	}
	m_search_error = true;
	return {};
}

template <typename T>
bool FindTVisitor<T>::is_path_found() const {
	return m_search.size() == m_search_match_count;
}

template class FindTVisitor<TorrentStringType>;
template class FindTVisitor<TorrentIntType>;
template class FindTVisitor<TorrentValue>;
} //namespace duck::detail
