/* Copyright 2025, Michele "King_DuckZ" Santullo
 * This file is part of ducktorrent.
 *
 * Ducktorrent is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Ducktorrent is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with ducktorrent.  If not, see <http://www.gnu.org/licenses/>.
 */

#pragma once

#include "ducktorrent/parser.hpp"
#include <string_view>
#include <cstddef>
#include <boost/variant/apply_visitor.hpp>
#include <vector>
#include <ciso646>
#include <type_traits>
#include <optional>

namespace duck {
namespace detail {
template <typename T>
class FindTVisitor : boost::static_visitor<T> {
public:
	typedef typename std::conditional<std::is_same<T, TorrentValue>::value,
		std::optional< std::reference_wrapper<const T> >,
		T
	>::type TRet;

	FindTVisitor (std::string_view search_path);

	void reset (const TorrentValue* curr_variant);
	bool has_error() const noexcept;
	TRet operator() (TorrentIntType value);
	TRet operator() (const TorrentStringType& value);
	TRet operator() (const std::vector<TorrentValue>&);
	TRet operator() (const std::map<TorrentStringType, TorrentValue>& value);

private:
	template <typename U>
	TRet scalar_operator_impl (const U& value);

	bool is_path_found() const;

	std::vector<std::string> m_search;
	const TorrentValue* m_curr_variant;
	std::size_t m_search_match_count{0};
	bool m_search_error{false};
};
} //namespace detail

typedef detail::FindTVisitor<TorrentStringType> FindStringVisitor;
typedef detail::FindTVisitor<TorrentIntType> FindIntVisitor;

template <typename T>
inline auto find_item(std::string_view path, const std::vector<TorrentValue>& values) {
	typedef typename detail::FindTVisitor<T>::TRet TRet;

	detail::FindTVisitor<T> visitor(path);
	for (const auto& value : values) {
		visitor.reset(&value);
		TRet found = boost::apply_visitor(visitor, value);
		if (not visitor.has_error())
			return found;
	}
	return TRet{};
}

template <typename T>
inline auto find_item(std::string_view path, const TorrentValue& value) {
	detail::FindTVisitor<T> visitor(path);
	visitor.reset(&value);
	return boost::apply_visitor(visitor, value);
}

template <typename T=TorrentIntType>
inline T find_int(std::string_view path, const std::vector<TorrentValue>& values) {
	return static_cast<T>(find_item<TorrentIntType>(path, values));
}

inline TorrentStringType find_string(std::string_view path, const std::vector<TorrentValue>& values) {
	return find_item<TorrentStringType>(path, values);
}

inline auto find_variant(std::string_view path, const std::vector<TorrentValue>& values) {
	return find_item<TorrentValue>(path, values);
}

template <typename T=TorrentIntType>
inline TorrentIntType find_int(std::string_view path, const TorrentValue& value) {
	return static_cast<T>(find_item<TorrentIntType>(path, value));
}

inline TorrentStringType find_string(std::string_view path, const TorrentValue& value) {
	return find_item<TorrentStringType>(path, value);
}

inline auto find_variant(std::string_view path, const TorrentValue& value) {
	return find_item<TorrentValue>(path, value);
}
} //namespace duck
